require 'open3'
require 'set'
require 'timeout'
require_relative 's3_uploader'

def self.safe_retry(cmd, description, error_message="command failed", raise_on_fail=true, log=true, tries=3, timeout=0)
    retry_count = 0
    last_try = false
    begin
        if (retry_count == tries - 1)
            last_try = true
        end

        status, output = system_safe(cmd, description, error_message, raise_on_fail=raise_on_fail, log=log, print_command=true, retry_exec=true, timeout=timeout, log_file_number=retry_count, last_try=last_try)
    rescue
        sleep 5
        retry if (retry_count += 1) < (tries)
    end

    if last_try and raise_on_fail and status != 0
        raise error_message
    else
        return status
    end
end

def self.get_sanitized_cmd(cmd)
    
    secret_fields = ["AWS_ACCESS_KEY_ID",
                     "AWS_SECRET_ACCESS_KEY",
                     "password",
                     "aws_access_key_id",
                     "aws_secret_access_key",
                     "aws_access_key",
                     "aws_secret_key",
                     "access_key",
                     "secret_key",
                     "account_access_key",
                     "account_secret_key",
                     "aws.accessKeyId",
                     "aws.secretKey",
                     "jwtsecret",
                     "master_access_key",
                     "master_secret_key",
                     "github_password"]
    
    cmd = cmd.gsub(/\s+/m, ' ').strip.split(" ")
    replace = cmd.grep(Regexp.union(secret_fields))
    unless replace.empty?
        replace = Set.new replace
        cmd.collect! {|e| (replace.include? e) ? '<secret>': e}
    end
    cmd = cmd.join(' ')
end

def self.pid_exists(pid)
    begin
        Process.getpgid(pid)
    rescue Errno::ESRCH
        return false
    end

    return true
end

def self.kill_pid(pid, timeout)
    puts "INFO: #{pid} timed out after #{timeout}, killing it"

    killpid = fork { Process.kill('INT', pid) }
    Process.detach(killpid)
    begin
        Process.waitpid(killpid)
    rescue Errno::ECHILD
        return true
    end

    return false
end

def self.execute_cmd(cmd, timeout, log_out, log_err)
    pid = spawn(cmd, :out=>log_out, :err=>log_err)
    begin
        return Timeout::timeout(timeout) do
            _, status = Process.wait2(pid)
            return status.success?
        end
    rescue Timeout::Error
        while pid_exists(pid)
            if kill_pid(pid, timeout)
                return false
            end
        end
    end
end

def self.log(level, message)
    date = Time.now.strftime("%Y-%m-%d %H:%M:%S")
    puts "#{date} #{level}: #{message}"
end

def self.logger(log, success, retry_exec, raise_on_fail, log_out, log_err, error_message, last_try)

    if log
        log("INFO", "Logs available here: #{log_err} and #{log_out}" )
    else
        log("INFO", "Logs unavailable due to user settings")
    end

    if not (success or retry_exec)
        log("ERROR", error_message)
        raise error_message
    end

    if not success and retry_exec
        if last_try and raise_on_fail
            log_level = "ERROR"
        else
            log_level = "WARNING"
        end

        if log
            log(log_level, "see #{log_out} and #{log_err} for details")
        end

        log(log_level, error_message)

        if not last_try
            raise error_message
        end
    end
end

def self.system_safe(cmd=cmd, cmd_description=cmd_description, error_message=error_message, raise_on_fail=true, log=true, print_command=true, retry_exec=false, timeout=0, log_file_number=1, last_try=true)
    start = Time.now.to_i
    
    cmd_sanitized = get_sanitized_cmd(cmd)
    if print_command
        log("INFO", "Executing command: #{cmd_sanitized}")
    else
        log("INFO", "Not logging command")
    end

    if log
        log_out = "logs/#{cmd_description}_out.log.#{log_file_number}"
        log_err = "logs/#{cmd_description}_err.log.#{log_file_number}"
    else
        log_out = "/dev/null"
        log_err = "/dev/null"
    end

    success = execute_cmd(cmd, timeout, log_out, log_err)

    elapsed = Time.now.to_i - start

    log("INFO", "#{cmd_description} took #{elapsed} seconds")

    file = File.open(log_out, "rb")
    out = file.read

    logger(log, success, retry_exec, raise_on_fail, log_out, log_err, error_message, last_try)

    return success
end

def self.system_unsafe(cmd)
	success = system(cmd)
	return success
end

def self.get_secret(key, subkey, log_identifier)
    secret = `vault read -field=#{subkey} #{key}`
    secret = secret.strip

    return secret
end

def self.put_secret(key, subkey_value_map, log_identifier)
	cmd = "vault write #{key} #{subkey_value_map}"
	system_safe(cmd, log_identifier, "Put secret failed", true, false)
end

def self.generate_aws_keys(key, log_identifier)
	cmd = "vault read #{key}"

	status, output = system_safe(cmd, log_identifier, "AWS key generation failed", true, false)

	cmd = "echo '#{output}' |grep access_key|awk '{print $2'}"
	status, account_access_key = system_safe(cmd, "get_generated_access_key", "Getting account access key failed", true, false, false)

	account_access_key = account_access_key.strip

	cmd = "echo '#{output}' |grep secret_key|awk '{print $2'}"
	status, account_secret_key = system_safe(cmd, "get_generated_access_key", "Getting account access key failed", true, false, false)
	account_secret_key = account_secret_key.strip

	cmd = "echo '#{output}' |grep lease_id|awk '{print $2'}"
	status, lease_id = system_safe(cmd, "get_generated_access_key", "Getting account access key failed", true, false, false)
	lease_id = lease_id.strip

	return account_access_key, account_secret_key, lease_id
end

def self.revoke_lease(lease_id, log_identifier)
	cmd = "vault revoke #{lease_id}"
	safe_retry(cmd, log_identifier, "Revoking lease failed", false, "ERROR", false)
end

def self.install_socotra_py_modules(pymodule, environment)
    cmd = "sudo pip list |grep #{pymodule}; if [ $? -eq 0 ];then /usr/bin/yes|sudo pip uninstall #{pymodule};fi"
    system_safe(cmd, "uninstall_#{pymodule}_py_module", "Failed to uninstall #{pymodule}")

    cmd = "/usr/bin/yes|sudo pip install --egg --process-dependency-links --allow-external mysql-connector-python --trusted-host socotra-pypi-euw1.s3-website-eu-west-1.amazonaws.com --extra-index-url=http://socotra-pypi-euw1.s3-website-eu-west-1.amazonaws.com/#{pymodule}/#{environment} #{pymodule}"
    system_safe(cmd, "install_#{pymodule}_py_module", "Failed to install #{pymodule}")
end

def self.sync_logs(repo, build_id)
    cmd = "cp -f console.log logs/"
    system_safe(cmd, "copy_console_log", "Failed to copy console log")

    upload_to_s3("logs", "socotra-builds/#{repo}/#{build_id}")
end
